@@ -14,6 +14,7 @@ int _gl_get_integer(int id) {
 void gen_textures(int count) {
 	Inline_Stack_Vars;
 	GLuint static_buf[16], *buf, i;
+	(void)items; /* squelch warning */
 
 	if (count < sizeof(static_buf)/sizeof(GLuint))
 		buf= static_buf;
@@ -57,6 +58,7 @@ void delete_textures(SV *first, ...) {
 void gen_buffers(int count) {
 	Inline_Stack_Vars;
 	GLuint static_buf[16], *buf, i;
+	(void)items; /* suppress warning */
 
 	if (count < sizeof(static_buf)/sizeof(GLuint))
 		buf= static_buf;
@@ -96,9 +98,11 @@ void delete_buffers(unsigned buf_id) {
 
 #endif
 #ifdef GL_VERSION_3_0
+
 void gen_vertex_arrays(int count) {
 	Inline_Stack_Vars;
 	GLuint static_buf[16], *buf, i;
+	(void)items; /* suppress warning */
 
 	if (count < sizeof(static_buf)/sizeof(GLuint))
 		buf= static_buf;
@@ -137,50 +141,125 @@ void delete_vertex_arrays(unsigned buf_id) {
 }
 #endif
 
-void _texture_load_rgb_square(HV *self, SV *mmap, int is_bgr) {
+/* This assumes it is being called as a method on a Texture object.
+ * Anything specific to the current option is passed as a parameter; anything about the
+ * format or configuration of the texture is passed via the object.
+ * The object is updated to match any new details about what was loaded into it.
+ */
+void _texture_load(HV *self, int level, int xoffset, int yoffset, int width, int height, int format, int type, SV *data_sv, int pitch) {
 	SV *sv;
-	void *data= SCALAR_REF_DATA(mmap);
-	int len= SCALAR_REF_LEN(mmap);
 	const char *ver;
-	int major, minor;
+	void *data;
+	int major, minor, tx_id, data_len, internal_fmt, pixel_size, target, need;
+	int known_format, has_alpha, default_internal_fmt, with_mipmaps;
+	GLint bound_pbo, orig_pix_align, orig_row_len, pix_align, row_len;
 	SV *tx_id_p=  _fetch_if_defined(self, "tx_id", 5);
 	SV *mipmap_p= _fetch_if_defined(self, "mipmap", 6);
 	SV *wrap_s_p= _fetch_if_defined(self, "wrap_s", 6);
 	SV *wrap_t_p= _fetch_if_defined(self, "wrap_t", 6);
 	SV *min_filter_p= _fetch_if_defined(self, "min_filter", 10);
 	SV *mag_filter_p= _fetch_if_defined(self, "mag_filter", 10);
-	int has_alpha= 0;
-	int dim= _dimension_from_filesize(len, &has_alpha);
-	int gl_fmt= is_bgr? ( has_alpha? GL_BGRA : GL_BGR )
-	                  : ( has_alpha? GL_RGBA : GL_RGB );
-	int gl_internal_fmt= has_alpha? GL_RGBA : GL_RGB;
-	
-	if (!tx_id_p) croak("tx_id must be initialized first");
-	
-	/* use mipmaps if the user set it to true, or if the min_filter uses a mipmap,
-	   and default in absence of any user prefs is true. */
-	int with_mipmaps= mipmap_p? SvTRUE(mipmap_p)
-		: !min_filter_p? 1
-		: SvIV(min_filter_p) == GL_NEAREST || SvIV(min_filter_p) == GL_LINEAR ? 0
-		: 1;
-	
-	if (!data || !len)
-		croak("Expected non-empty scalar-ref pixel buffer");
-	
-	/* Ensure the OpenGL context is initialized */
-	//call_pv("OpenGL::Sandbox::_ensure_context", G_VOID | G_NOARGS | G_EVAL);
-	//if (SvTRUE(ERRSV))
-	//	croak(NULL);
+	SV *internal_p= _fetch_if_defined(self, "internal_format", 15);
+	SV *target_p= _fetch_if_defined(self, "target", 6);
 	
 	/* Mipmap strategy depends on version of GL.
 	   Supposedly this GetString is more compatible than GetInteger(GL_VERSION_MAJOR)
 	*/
 	ver= (const char *) glGetString(GL_VERSION);
 	if (!ver || sscanf(ver, "%d.%d", &major, &minor) != 2)
-		croak("Can't get GL_VERSION");
+		carp_croak("Can't get GL_VERSION");
 	
-	/* Bind texture */
-	glBindTexture(GL_TEXTURE_2D, SvIV(tx_id_p));
+	/* TODO: support things other than GL_TEXTURE_2D. */
+	if (target_p && SvIV(target_p) != GL_TEXTURE_2D)
+		carp_croak("Texture targets other than GL_TEXTURE_2D not yet supported");
+	target= GL_TEXTURE_2D;
+
+	known_format= _get_format_info(format, NULL, &has_alpha, &default_internal_fmt);
+	pixel_size= _get_pixel_size(format, type);
+
+	/* Data argument is hard to validate.  It should normally be a scalar ref, but could also be NULL to create
+	 * texture storage without loading, and when using PBOs could also be a plain integer offset within the PBO */
+	#ifdef GL_PIXEL_UNPACK_BUFFER_BINDING
+	bound_pbo= 0;
+	if (major >= 2)
+		glGetIntegerv(GL_PIXEL_UNPACK_BUFFER_BINDING, &bound_pbo);
+	if (bound_pbo) {
+		if (SvOK(data_sv) && !(SvIOK(data_sv) || SvUOK(data_sv)))
+			carp_croak("PBO for UNPACK is active; pixel 'data' must be a numeric offset, or undef");
+		data= (void*) SvUV(data_sv);
+	}
+	else
+	#endif
+	{
+		if (SvROK(data_sv)) {
+			data= SCALAR_REF_DATA(data_sv); /* NULL is permitted for using PBOs or initializing storage without loading it */
+			data_len= SCALAR_REF_LEN(data_sv);
+			need= width * height * pixel_size;
+			if (need > data_len)
+				carp_croak("Require at least %d bytes of pixel data (got %d)", need, data_len);
+		}
+		else if (!SvOK(data_sv) || SvIV(data_sv) == 0) {
+			if (xoffset || yoffset) carp_croak("Can't use NULL pixel data when specifying a sub-image");
+			data= NULL;
+			data_len= 0;
+		}
+		else
+			carp_croak("Expected scalar-ref %sfor data argument", (xoffset || yoffset)? "":"or undef ");
+	}
+	
+	/* TODO: support OpenGL 4.5 which doesn't need to bind to anything */
+	if (!tx_id_p || !(tx_id= SvUV(tx_id_p)))
+		croak("tx_id must be initialized first");
+	glBindTexture(target, tx_id);
+	
+	if (pitch) {
+		/* OpenGL doesn't do row length in bytes, it does it in pixels. This is not helpful. */
+		if (!pixel_size)
+			croak("Don't know how to apply 'pitch' to pixels of format=%d type=%d", format, type);
+		switch (pitch > 0 ? pitch % pixel_size : -1) {
+		case 0: pix_align= 1; row_len= pitch / pixel_size; break;
+		case 1: if ((pitch & 1) == 0) { pix_align= 2; row_len= pitch / pixel_size; break; }
+		case 2:
+		case 3: if ((pitch & 3) == 0) { pix_align= 4; row_len= pitch / pixel_size; break; }
+		case 4:
+		case 5:
+		case 6:
+		case 7: if ((pitch & 7) == 0) { pix_align= 8; row_len= pitch / pixel_size; break; }
+		default:
+			croak("Unsupported buffer pitch %d for pixel size %d", pitch, pixel_size);
+		}
+		glGetIntegerv(GL_UNPACK_ALIGNMENT, &orig_pix_align);
+		glGetIntegerv(GL_UNPACK_ROW_LENGTH, &orig_row_len);
+	}
+	
+	/* If xoffset or yoffset are nonzero, then this requires the texture to be loaded already,
+	 * and the internal format and mipmap and etc is irrelevant. */
+	if (xoffset || yoffset) {
+		if (pitch) {
+			glPixelStorei(GL_UNPACK_ROW_LENGTH, row_len);
+			glPixelStorei(GL_UNPACK_ALIGNMENT, pix_align);
+		}
+		glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, data);
+		if (pitch) {
+			glPixelStorei(GL_UNPACK_ROW_LENGTH, orig_row_len);
+			glPixelStorei(GL_UNPACK_ALIGNMENT, orig_pix_align);
+		}
+		return;
+	}
+	/* Else we are defining the storage for the texture and more things need considered.
+	 * Also the texture object should be updated with the result of the calculations below. */
+	
+	if (internal_p) internal_fmt= SvIV(internal_p);
+	else if (known_format) internal_fmt= default_internal_fmt;
+	else carp_croak("No default internal_format for given format %d; must be specified", format);
+	
+	/* use mipmaps if the user set it to true, or if the min_filter uses a mipmap,
+	   and default in absence of any user prefs is true. But not if the user has specified 'level' */
+	with_mipmaps= level? 0
+		: mipmap_p? SvTRUE(mipmap_p)
+		: !min_filter_p? 1
+		: SvIV(min_filter_p) == GL_NEAREST || SvIV(min_filter_p) == GL_LINEAR ? 0
+		: 1;
 	
 	if (with_mipmaps) {
 		if (major < 3) {
@@ -190,7 +269,7 @@ void _texture_load_rgb_square(HV *self, SV *mmap, int is_bgr) {
 			if (min_filter_p)
 				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SvIV(min_filter_p));
 		}
-	} else {
+	} else if (!level) {
 		if (mag_filter_p)
 			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, SvIV(mag_filter_p));
 		/* this one needs overridden even if user didn't request it, because default uses mipmaps */
@@ -199,9 +278,17 @@ void _texture_load_rgb_square(HV *self, SV *mmap, int is_bgr) {
 		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
 		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
 	}
-	glTexImage2D(GL_TEXTURE_2D, 0, gl_internal_fmt, dim, dim, 0, gl_fmt, GL_UNSIGNED_BYTE, data);
+	if (pitch) {
+		glPixelStorei(GL_UNPACK_ROW_LENGTH, row_len);
+		glPixelStorei(GL_UNPACK_ALIGNMENT, pix_align);
+	}
+	glTexImage2D(target, level, internal_fmt, width, height, 0, format, type, data);
+	if (pitch) {
+		glPixelStorei(GL_UNPACK_ROW_LENGTH, orig_row_len);
+		glPixelStorei(GL_UNPACK_ALIGNMENT, orig_pix_align);
+	}
 	if (with_mipmaps && major >= 3) {
-		/* glEnable(GL_TEXTURE_2D); /* correct bug in ATI, accoridng to Khronos FAQ */
+		/* glEnable(GL_TEXTURE_2D);  correct bug in ATI, accoridng to Khronos FAQ */
 		glGenerateMipmap(GL_TEXTURE_2D);
 		/* examples show setting these after mipmap generation.  Does it matter? */
 		if (mag_filter_p)
@@ -209,16 +296,20 @@ void _texture_load_rgb_square(HV *self, SV *mmap, int is_bgr) {
 		if (min_filter_p)
 			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SvIV(min_filter_p));
 	}
-	if (wrap_s_p)
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, SvIV(wrap_s_p));
-	if (wrap_t_p)
-		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, SvIV(wrap_t_p));
+	if (!level) {
+		if (wrap_s_p)
+			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, SvIV(wrap_s_p));
+		if (wrap_t_p)
+			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, SvIV(wrap_t_p));
+	}
 
 	/* update attributes */
-	if (!hv_store(self, "width",     5, sv=newSViv(dim), 0)
-	 || !hv_store(self, "height",    6, sv=newSViv(dim), 0)
-	 || !hv_store(self, "pow2_size", 9, sv=newSViv(dim), 0)
-	 || !hv_store(self, "has_alpha", 9, sv=newSViv(has_alpha? 1 : 0), 0)
+	if (!hv_store(self, "width",            5, sv=newSViv(width), 0)
+	 || !hv_store(self, "height",           6, sv=newSViv(height), 0)
+	 || (known_format &&
+	    !hv_store(self, "has_alpha",        9, sv=newSViv(has_alpha? 1 : 0), 0))
+	 || !hv_store(self, "internal_format", 15, sv=newSViv(internal_fmt), 0)
+	 || !hv_store(self, "loaded",           6, sv=newSViv(1), 0)
 	) {
 		if (sv) sv_2mortal(sv);
 		croak("Can't store results in supplied hash");
@@ -237,51 +328,16 @@ int _round_up_pow2(long dim) {
 	return ++dim;
 }
 
-SV* _img_rescale_to_pow2_square(int width, int height, int has_alpha, int want_bgr, SV *sref) {
-	struct SwsContext *sws= NULL;
-	SV *ret= NULL;
-	void *data= SCALAR_REF_DATA(sref);
-	int len= SCALAR_REF_LEN(sref);
-	int px_size= has_alpha? 4 : 3;
-	int dim= _round_up_pow2(width > height? width : height);
-	const uint8_t *src_planes[4]= { data,0,0,0 };
-	int src_stride[4]= { width*px_size,0,0,0 };
-	uint8_t *dst_planes[4]= { 0,0,0,0 };
-	int dst_stride[4]= { dim*px_size,0,0,0 };
-	
-	if (!data || !len)
-		croak("Expected non-empty scalar-ref pixel buffer");
-	if (width * height * px_size != len)
-		croak("Size of scalar ref disagrees with rectangle dimensions: %d * %d * %d != %d",
-			width, height, px_size, len);
-	
-	/* rescale to square */
-	sws= sws_getCachedContext(sws, width, height, has_alpha? AV_PIX_FMT_RGBA : AV_PIX_FMT_RGB24,
-		dim, dim, want_bgr? (has_alpha? AV_PIX_FMT_BGRA : AV_PIX_FMT_BGR24) : (has_alpha? AV_PIX_FMT_RGBA : AV_PIX_FMT_RGB24),
-		SWS_BICUBIC, NULL, NULL, NULL);
-	if (!sws)
-		croak("can't initialize resize context");
-	
-	/* allocate a "mortal" scalar into which we write the new image */
-	ret= sv_2mortal(newSV(dim*dim*px_size));
-	sv_setpvn(ret, "", 0);
-	SvGROW(ret, dim*dim*px_size);
-	SvCUR_set(ret, dim*dim*px_size);
-	dst_planes[0]= (uint8_t*) SvPVX(ret);
-	
-	/* perform the rescale */
-	sws_scale(sws, src_planes, src_stride, 0, height, dst_planes, dst_stride);
-	sws_freeContext(sws);
-	
-	/* return a ref to the scalar, to avoid making a copy */
-	return newRV_inc(ret);
+/* for diagnosing changes to memory-mapped scalar refs */
+intptr_t _get_scalarref_pv(SV *sv) {
+	return (intptr_t) SCALAR_REF_DATA(sv);
 }
 
 void _img_rgb_to_bgr(SV *sref, int has_alpha) {
 	char *p= SCALAR_REF_DATA(sref), *last, c;
 	int px_size= has_alpha? 4 : 3;
 	int len= SCALAR_REF_LEN(sref);
-	if (!p || len < px_size) croak("Expected non-empty scalar-ref pixel buffer");
+	if (!p || len < px_size) carp_croak("Expected non-empty scalar-ref pixel buffer");
 	last= p + len - px_size;
 	while (p <= last) {
 		c= p[0]; p[0]= p[2]; p[2]= c;
@@ -320,7 +376,7 @@ void load_buffer_data(int target, SV *size_sv, SV *data_sv, SV *usage_sv) {
 	char *data= NULL;
 	_get_buffer_from_sv(data_sv, &data, &data_size);
 	size= (size_sv && SvOK(size_sv))? SvUV(size_sv) : data_size;
-	if (data_size < size) croak("Data not long enough (%d bytes, you requested %d)", data_size, size);
+	if (data_size < size) carp_croak("Data not long enough (%d bytes, you requested %d)", data_size, size);
 	glBufferData(target, size, data, usage);
 }
 
@@ -330,15 +386,153 @@ void load_buffer_sub_data(int target, long offset, SV *size_sv, SV *data_sv, SV
 	_get_buffer_from_sv(data_sv, &data, &data_size);
 	if (data_offset_sv && SvOK(data_offset_sv)) {
 		data_offset= SvUV(data_offset_sv);
-		if (data_offset > data_size) croak("Invalid data offset (%d exceeds data length %d)", data_offset, data_size);
+		if (data_offset > data_size) carp_croak("Invalid data offset (%ld exceeds data length %ld)", data_offset, data_size);
 		data += data_offset;
 		data_size -= data_offset;
 	}
 	size= (size_sv && SvOK(size_sv))? SvUV(size_sv) : data_size;
-	if (data_size < size) croak("Data not long enough (%d bytes, you requested %d)", data_size, size);
+	if (data_size < size) carp_croak("Data not long enough (%ld bytes, you requested %ld)", data_size, size);
 	glBufferSubData(target, offset, size, data);
 }
 
+SV *mmap_buffer(int buffer_id, SV *target_sv, SV *access_sv, SV *offset_sv, SV *length_sv) {
+	int gl_maj= 0, gl_min= 0;
+	int access= 0, access_r= 0, access_w= 0, mode;
+	GLint actual_size= 0, target;
+	STRLEN len;
+	unsigned offset, length;
+	const char* access_pv;
+	void *addr;
+	SV *sv;
+	buffer_scalar_callback_data_t cbdata;
+	
+	/* OpenGL 2.0 only has MapBuffer, 3.0 has MapBufferRange (needed for access flags)
+	 * and OpenGL 4.5 has MapNamedBufferRange needed to avoid binding the buffer first
+	 */
+	glGetIntegerv(GL_MAJOR_VERSION, &gl_maj);
+	glGetIntegerv(GL_MINOR_VERSION, &gl_min);
+
+	/* 'access' can be given as a symbolic string, or as an integer.  If omitted, assume "r+".
+	 * OpenGL < 3.0 won't even have the constants available for GL_MAP_*, so the symbolic
+	 * constants allow generic access to the API without worrying about that.
+	 */
+	if (!SvOK(access_sv)) {
+		access_r= 1;
+		access_w= 1;
+		#ifdef GL_VERSION_3_0
+		access= GL_MAP_READ_BIT | GL_MAP_WRITE_BIT;
+		#endif
+	}
+	#ifdef GL_VERSION_3_0
+	else if (sv_contains_integer(access_sv)) {
+		access= SvIV(access_sv);
+		access_r= (access & GL_MAP_READ_BIT)? 1 : 0;
+		access_w= (access & GL_MAP_WRITE_BIT)? 1 : 0;
+	}
+	#endif
+	else {
+		access_pv= SvPV(access_sv, len);
+		if (len < 1) carp_croak("Invalid symbolic access notation \"\"");
+		while (*access_pv) {
+			switch (*access_pv++) {
+			case '+': access_r= 1, access_w= 1; break;
+			case 'r': access_r= 1; break;
+			case 'w': access_w= 1; break;
+			default: carp_croak("Invalid symbolic access notation '%c' in '%s'", access_pv[-1], SvPV_nolen(access_sv));
+			}
+		}
+		#ifdef GL_VERSION_3_0
+		access= (access_r? GL_MAP_READ_BIT : 0) | (access_w? GL_MAP_WRITE_BIT : 0);
+		#endif
+	}
+	
+	offset= SvOK(offset_sv)? SvUV(offset_sv) : 0;
+	length= SvOK(length_sv)? SvUV(length_sv) : 0;
+
+	/* OpenGL 4.5 can look up size and map buffer without binding first */
+	#ifdef GL_VERSION_4_5
+	if (gl_maj >= 4 && gl_min >= 5) {
+		glGetNamedBufferParameteriv(buffer_id, GL_BUFFER_SIZE, &actual_size);
+	}
+	else
+	#endif
+	{
+		if (!SvOK(target_sv)) carp_croak("Require GL buffer target on OpenGL < 4.5");
+		target= SvIV(target_sv);
+		glBindBuffer(target, buffer_id);
+		glGetBufferParameteriv(target, GL_BUFFER_SIZE, &actual_size);
+	}
+
+	/* Make sure the length and offset make sense */
+	if (!actual_size) carp_croak("Cannot mem-map buffer until storage has been allocated");
+	if (offset > actual_size) carp_croak("Offset %d exceeds actual buffer size %d", offset, actual_size);
+	if (offset+length > actual_size) carp_croak("Length %d exceeds actual buffer size %d", length, actual_size);
+	if (!length) length= actual_size - offset;
+	if (!length) carp_croak("Cannot map zero bytes of a buffer"); 
+
+	/* OpenGL 4.5 can look up size and map buffer without binding first */
+	#ifdef GL_VERSION_4_5
+	if (gl_maj >= 4 && gl_min >= 5) {
+		if (!(addr= glMapNamedBufferRange(buffer_id, offset, length, access)))
+			carp_croak("glMapNamedBufferRange failed");
+	}
+	else
+	#endif
+	{
+		/* OpenGL 3.0 is required for BufferRange, else fall back to mapping whole thing. */
+		#ifdef GL_VERSION_3_0
+		if (gl_maj >= 3) {
+			if (!(addr= glMapBufferRange(target, offset, length, access)))
+				carp_croak("glMapBufferRange failed");
+		}
+		else
+		#endif
+		{
+			if (!access_r && !access_w)
+				carp_croak("Must specify read or write access");
+			mode= access_r && access_w? GL_READ_WRITE
+				  : access_r? GL_READ_ONLY
+				  : GL_WRITE_ONLY;
+			if (!(addr= glMapBuffer(target, mode)))
+				carp_croak("glMapBuffer failed");
+			/* OpenGL mapped all of it, but we can just pretend we did a sub-range */
+			addr= (void*) (((char*) addr) + offset);
+		}
+	}
+
+	/* at this point, have buffer mapped and know length */
+	sv= sv_2mortal(newRV_noinc((SV*)newSV(0)));
+	buffer_scalar_wrap(SvRV(sv), addr, length, access_w? 0 : BUFFER_SCALAR_READONLY, cbdata, NULL);
+	sv_bless(sv, gv_stashpv("OpenGL::Sandbox::MMap", GV_ADD));
+	return SvREFCNT_inc(sv);
+}
+
+int unmap_buffer(int buffer_id, SV *target_sv, SV *memmap) {
+	int gl_maj= 0, gl_min= 0, target;
+	if (sv_isa(memmap, "OpenGL::Sandbox::MMap")) {
+		buffer_scalar_unwrap(SvRV(memmap));
+		sv_setsv(memmap, &PL_sv_undef);
+	}
+	/* OpenGL 4.5 has UnmapNamedBuffer, else we have to bind the buffer first */
+	#ifdef GL_VERSION_4_5
+	glGetIntegerv(GL_MAJOR_VERSION, &gl_maj);
+	if (gl_maj >= 4) {
+		glGetIntegerv(GL_MINOR_VERSION, &gl_min);
+		if (gl_min >= 5) {
+			glUnmapNamedBuffer(buffer_id);
+			return 1;
+		}
+	}
+	#else
+	(void) gl_maj; (void) gl_min; /* stupid warnings */
+	#endif
+	if (!SvOK(target_sv) || !(target= SvIV(target_sv)))
+		carp_croak("Must specify buffer target for OpenGL < 4.5");
+	glBindBuffer(target, buffer_id);
+	glUnmapBuffer(target);
+	return 1;
+}
+
 const char* get_glsl_type_name(int type) {
 	switch (type) {
 	case GL_BOOL:              return "bool";
@@ -417,8 +611,8 @@ SV * get_program_uniforms(unsigned program) {
 void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 	Inline_Stack_Vars;
 	SV **entry, *s;
-	int type= 0, component_type, size= 0, loc= 0, components= 0, buf_req= 0, i, cur_prog, arg_i, arg_lim, dest_i;
-	unsigned long buf_size;
+	int type= 0, component_type, size= 0, loc= 0, components= 0, i, cur_prog, arg_i, arg_lim, dest_i;
+	unsigned long buf_size, buf_req= 0;
 	char static_buf[ 8 * 16 ], *buf= NULL;
 	AV *info= NULL;
 	
@@ -429,7 +623,7 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 		glGetIntegerv(GL_MAJOR_VERSION, &i);
 		if (i < 4)
 		#endif
-			croak("Can't set uniforms for program other than the current (unless GL >= 4.1)");
+			carp_croak("Can't set uniforms for program other than the current (unless GL >= 4.1)");
 	}
 
 	/* Lazy-build the uniform cache */
@@ -440,19 +634,19 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 	/* Find uniform details by name */
 	entry= hv_fetch((HV*) SvRV(uniform_cache), name, strlen(name), 0);
 	if (!entry || !*entry || !SvROK(*entry))
-		croak("No active uniform '%s' in program %d", name, program);
+		carp_croak("No active uniform '%s' in program %d", name, program);
 	if (SvTYPE(SvRV(*entry)) != SVt_PVAV || av_len(info= (AV*) SvRV(*entry)) < 3)
-		croak("Invalid uniform info record for %s", name);
+		carp_croak("Invalid uniform info record for %s", name);
 
 	/* Validate the uniform metadata */
 	entry= av_fetch(info, 1, 0);
-	if (!entry || !*entry || !SvIOK(*entry)) croak("Invalid uniform info record for %s", name);
+	if (!entry || !*entry || !SvIOK(*entry)) carp_croak("Invalid uniform info record for %s", name);
 	loc= SvIV(*entry);
 	entry= av_fetch(info, 2, 0);
-	if (!entry || !*entry || !SvIOK(*entry)) croak("Invalid uniform info record for %s", name);
+	if (!entry || !*entry || !SvIOK(*entry)) carp_croak("Invalid uniform info record for %s", name);
 	type= SvIV(*entry);
 	entry= av_fetch(info, 3, 0);
-	if (!entry || !*entry || !SvIOK(*entry)) croak("Invalid uniform info record for %s", name);
+	if (!entry || !*entry || !SvIOK(*entry)) carp_croak("Invalid uniform info record for %s", name);
 	size= SvIV(*entry);
 
 	/* Determine how many and what type of arguments we want based on type */
@@ -505,7 +699,7 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 		break;
 	#endif
 	default:
-		croak("Unimplemented type %d for uniform %s", type, name);
+		carp_croak("Unimplemented type %d for uniform %s", type, name);
 	}
 
 	/* If there is only one argument, and it is a ref, and not an arrayref (those get handled below)
@@ -516,9 +710,10 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 		if (SvROK(s) && SvTYPE(SvRV(s)) != SVt_PVAV) {
 			_get_buffer_from_sv(s, &buf, &buf_size);
 			if (!buf || !buf_size)
-				croak("Don't know how to extract values/buffer from %s", SvPV_nolen(s));
+				carp_croak("Don't know how to extract values/buffer from %s", SvPV_nolen(s));
 			if (buf_size < buf_req)
-				croak("Uniform %s is type %s, requiring packed data of at least %d bytes (got %d)", name, get_glsl_type_name(type), buf_req, buf_size);
+				carp_croak("Uniform %s is type %s, requiring packed data of at least %ld bytes (got %ld)",
+					name, get_glsl_type_name(type), buf_req, buf_size);
 		}
 	}
 	/* If not given a packed buffer, recursively iterate the arguments and pack it into one of our own */
@@ -533,7 +728,7 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 		for (arg_i= 3, arg_lim= Inline_Stack_Items; arg_i < arg_lim; ++arg_i)
 			_recursive_pack(buf, &dest_i, components*size, component_type, Inline_Stack_Item(arg_i));
 		if (dest_i != components*size)
-			croak("Uniform %s is type %s, requiring %d values (got %d)", name, get_glsl_type_name(type), components*size, dest_i);
+			carp_croak("Uniform %s is type %s, requiring %d values (got %d)", name, get_glsl_type_name(type), components*size, dest_i);
 	}
 
 	/* Finally, call glUniform depending on the type */
@@ -583,7 +778,7 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 	case GL_DOUBLE_MAT3x4: glUniformMatrix3x4dv(loc, size, 0, (GLdouble*) buf); break;
 	case GL_DOUBLE_MAT4x3: glUniformMatrix4x3dv(loc, size, 0, (GLdouble*) buf); break;
 	#endif
-	default: croak("Unimplemented type %d for uniform %s", type, name);
+	default: carp_croak("Unimplemented type %d for uniform %s", type, name);
 	}
 	#if 0
 	//#ifdef GL_VERSION_4_1
@@ -623,7 +818,7 @@ void set_uniform(unsigned program, SV* uniform_cache, const char *name, ...) {
 		case GL_DOUBLE_MAT4x2: glProgramUniformMatrix4x2dv(program, loc, size, 0, (GLdouble*) buf); break;
 		case GL_DOUBLE_MAT3x4: glProgramUniformMatrix3x4dv(program, loc, size, 0, (GLdouble*) buf); break;
 		case GL_DOUBLE_MAT4x3: glProgramUniformMatrix4x3dv(program, loc, size, 0, (GLdouble*) buf); break;
-		default: croak("Unimplemented type %d for uniform %s", type, name);
+		default: carp_croak("Unimplemented type %d for uniform %s", type, name);
 		}
 	}
 	#endif
